@@ -9,6 +9,10 @@ import re
 import six
 import sys
 
+# Package version
+VERSION = "0.9.9"
+
+
 # ------------------------------------ Python 2/3 compatibility challenges  ----
 # Python 3 no longer has a basestring type, so....
 try:
@@ -23,7 +27,7 @@ try:
 
     def str_(s):
         """
-        Return string with correct encoding
+        Return byte string with correct encoding
         """
         if type(s) == unicode:
             return s.encode('utf-8')
@@ -69,18 +73,18 @@ def to_basestring(s):
 logger = logging.getLogger(__name__)
 if not logging.getLogger().handlers:
     handler = logging.StreamHandler()
-    formatter = logging.Formatter('{name} {levelname} {message}')
+    formatter = logging.Formatter('%(name)s %(levelname)s %(message)s')
     handler.setFormatter(formatter)
     logger.addHandler(handler)
 logger.setLevel(logging.ERROR)  # Log errors
 DEBUG = False  # Don't waste time on debug calls
 
 # ----------------------------------- Constants --------------------------------
-CR     = '\r'
-LF     = '\n'
-CRLF   = CR + LF
-SPACE  = ' '
-TAB    = '\t'
+CR = '\r'
+LF = '\n'
+CRLF = CR + LF
+SPACE = ' '
+TAB = '\t'
 SPACEORTAB = SPACE + TAB
 
 # --------------------------------- Main classes -------------------------------
@@ -147,7 +151,7 @@ class VBase(object):
         """
         parentBehavior = self.parentBehavior
         if parentBehavior is not None:
-            knownChildTup = parentBehavior.knownChildren.get(self.name, None)
+            knownChildTup = parentBehavior.knownChildren.get(self.name)
             if knownChildTup is not None:
                 behavior = getBehavior(self.name, knownChildTup[2])
                 if behavior is not None:
@@ -239,7 +243,7 @@ class VBase(object):
         """
         pass
 
-    def serialize(self, buf=None, lineLength=75, validate=True, behavior=None):
+    def serialize(self, buf=None, lineLength=75, validate=True, behavior=None, *args, **kwargs):
         """
         Serialize to buf if it exists, otherwise return a string.
 
@@ -251,7 +255,7 @@ class VBase(object):
         if behavior:
             if DEBUG:
                 logger.debug("serializing {0!s} with behavior {1!s}".format(self.name, behavior))
-            return behavior.serialize(self, buf, lineLength, validate)
+            return behavior.serialize(self, buf, lineLength, validate, *args, **kwargs)
         else:
             if DEBUG:
                 logger.debug("serializing {0!s} without behavior".format(self.name))
@@ -323,12 +327,12 @@ class ContentLine(VBase):
         list(map(updateTable, params))
 
         qp = False
-        if 'ENCODING' in self.params:
-            if 'QUOTED-PRINTABLE' in self.params['ENCODING']:
-                qp = True
-                self.params['ENCODING'].remove('QUOTED-PRINTABLE')
-                if len(self.params['ENCODING']) == 0:
-                    del self.params['ENCODING']
+        if 'ENCODING' in self.params and \
+                'QUOTED-PRINTABLE' in self.params['ENCODING']:
+            qp = True
+            self.params['ENCODING'].remove('QUOTED-PRINTABLE')
+            if len(self.params['ENCODING']) == 0:
+                del self.params['ENCODING']
         if 'QUOTED-PRINTABLE' in self.singletonparams:
             qp = True
             self.singletonparams.remove('QUOTED-PRINTABLE')
@@ -339,11 +343,11 @@ class ContentLine(VBase):
                 if 'CHARSET' in self.params:
                     self.value = codecs.decode(self.value.encode("utf-8"), "quoted-printable").decode(self.params['CHARSET'][0])
                 else:
-                    self.value = codecs.decode(self.value.encode("utf-8"), "quoted-printable").decode()
+                    self.value = codecs.decode(self.value.encode("utf-8"), "quoted-printable").decode('utf-8')
 
     @classmethod
-    def duplicate(clz, copyit):
-        newcopy = clz('', {}, '')
+    def duplicate(cls, copyit):
+        newcopy = cls('', {}, '')
         newcopy.copy(copyit)
         return newcopy
 
@@ -360,8 +364,7 @@ class ContentLine(VBase):
 
     def __eq__(self, other):
         try:
-            return (self.name == other.name) and (self.params == other.params)\
-                   and (self.value == other.value)
+            return (self.name == other.name) and (self.params == other.params) and (self.value == other.value)
         except Exception:
             return False
 
@@ -428,11 +431,17 @@ class ContentLine(VBase):
         return v
 
     def __str__(self):
-        return "<{0}{1}{2}>".format(self.name, self.params, self.valueRepr())
+        try:
+            return "<{0}{1}{2}>".format(self.name, self.params, self.valueRepr())
+        except UnicodeEncodeError as e:
+            return "<{0}{1}{2}>".format(self.name, self.params, self.valueRepr().encode('utf-8'))
 
     def __repr__(self):
         return self.__str__()
 
+    def __unicode__(self):
+        return u"<{0}{1}{2}>".format(self.name, self.params, self.valueRepr())
+
     def prettyPrint(self, level=0, tabwidth=3):
         pre = ' ' * level * tabwidth
         print(pre, self.name + ":", self.valueRepr())
@@ -473,8 +482,8 @@ class Component(VBase):
         self.autoBehavior()
 
     @classmethod
-    def duplicate(clz, copyit):
-        newcopy = clz()
+    def duplicate(cls, copyit):
+        newcopy = cls()
         newcopy.copy(copyit)
         return newcopy
 
@@ -600,9 +609,9 @@ class Component(VBase):
                 obj = obj.transformToNative()
             except (KeyError, AttributeError):
                 obj = ContentLine(objOrName, [], '', group)
-            if obj.behavior is None and self.behavior is not None:
-                if isinstance(obj, ContentLine):
-                    obj.behavior = self.behavior.defaultBehavior
+            if obj.behavior is None and self.behavior is not None and \
+                    isinstance(obj, ContentLine):
+                obj.behavior = self.behavior.defaultBehavior
         self.contents.setdefault(obj.name.lower(), []).append(obj)
         return obj
 
@@ -728,7 +737,7 @@ patterns = {}
 
 # Note that underscore is not legal for names, it's included because
 # Lotus Notes uses it
-patterns['name'] = '[a-zA-Z0-9\-_]+'
+patterns['name'] = '[a-zA-Z0-9_-]+'   # 1*(ALPHA / DIGIT / "-")
 patterns['safe_char'] = '[^";:,]'
 patterns['qsafe_char'] = '[^"]'
 
@@ -769,7 +778,7 @@ patterns['params_grouped'] = r"""
 # get a full content line, break it up into group, name, parameters, and value
 patterns['line'] = r"""
 ^ ((?P<group> {name!s})\.)?(?P<name> {name!s}) # name group
-  (?P<params> (?: {param!s} )* )               # params group (may be empty)
+  (?P<params> ;?(?: {param!s} )* )               # params group (may be empty)
 : (?P<value> .* )$                             # value group
 """.format(**patterns)
 
@@ -946,7 +955,7 @@ def foldOneLine(outbuf, input, lineLength=75):
             outbuf.write(bytes(input, 'UTF-8'))
         except Exception:
             # fall back on py2 syntax
-            outbuf.write(str_(input))
+            outbuf.write(input)
 
     else:
         # Look for valid utf8 range and write that out
@@ -1013,12 +1022,18 @@ def defaultSerialize(obj, buf, lineLength):
 
         if obj.group is not None:
             s.write(obj.group + '.')
-        s.write(obj.name.upper())
+        s.write(str_(obj.name.upper()))
         keys = sorted(obj.params.keys())
         for key in keys:
             paramstr = ','.join(dquoteEscape(p) for p in obj.params[key])
-            s.write(";{0}={1}".format(key, paramstr))
-        s.write(":{0}".format(str_(obj.value)))
+            try:
+                s.write(";{0}={1}".format(key, paramstr))
+            except (UnicodeDecodeError, UnicodeEncodeError):
+                s.write(";{0}={1}".format(key, paramstr.encode('utf-8')))
+        try:
+            s.write(":{0}".format(obj.value))
+        except (UnicodeDecodeError, UnicodeEncodeError):
+            s.write(":{0}".format(obj.value.encode('utf-8')))
         if obj.behavior and not startedEncoded:
             obj.behavior.decode(obj)
         foldOneLine(outbuf, s.getvalue(), lineLength)
@@ -1067,7 +1082,7 @@ def readComponents(streamOrString, validate=False, transform=True,
     Generate one Component at a time from a stream.
     """
     if isinstance(streamOrString, basestring):
-        stream = six.StringIO(str_(streamOrString))
+        stream = six.StringIO(streamOrString)
     else:
         stream = streamOrString
 
@@ -1205,5 +1220,5 @@ def newFromBehavior(name, id=None):
 
 # --------------------------- Helper function ----------------------------------
 def backslashEscape(s):
-    s = s.replace("\\", "\\\\").replace(";", "\;").replace(",", "\,")
+    s = s.replace("\\", "\\\\").replace(";", "\\;").replace(",", "\\,")
     return s.replace("\r\n", "\\n").replace("\n", "\\n").replace("\r", "\\n")
